unit dnssync;

interface
  uses
    dnscore,
    binipstuff;

  function forwardlookup(name:string;timeout:integer):tbinip;

var
  dnssyncserver:string;


implementation
  uses
    {$ifdef VER1_0}
      linux,
    {$else}
      baseunix,unix,
    {$endif}
    sockets,
    fd_utils,
    sysutils;
{$i unixstuff.inc}
{$i ltimevalstuff.inc}

var
  fd:integer;
  state:tdnsstate;

function sendquery(const packet:tdnspacket;len:integer):boolean;
var
  a:integer;
  addr       : string;
  port       : string;
  inaddr     : TInetSockAddr;

begin
{  writeln('sendquery ',decodename(state.packet,state.packetlen,12,0,a),' ',state.requesttype);}
  result := false;
  if len = 0 then exit; {no packet}

  if dnssyncserver <> '' then addr := dnssyncserver else addr := getunixnameservercached;
  port := '53';

  inAddr.family:=AF_INET;
  inAddr.port:=htons(strtointdef(port,0));
  inAddr.addr:=htonl(longip(addr));

  sendto(fd,packet,len,0,inaddr,sizeof(inaddr));
  result := true;
end;

procedure setupsocket;
var
  inAddrtemp : TInetSockAddr;
begin
  if fd > 0 then exit;

  fd := Socket(AF_INET,SOCK_DGRAM,0);
  inAddrtemp.family:=AF_INET;
  inAddrtemp.port:=0;
  inAddrtemp.addr:=0;{htonl(longip('0.0.0.0'));}
  If Not Bind(fd,inAddrtemp,SizeOf(inAddrtemp)) Then begin
    raise Exception.create('unable to bind');
  end;
end;

procedure resolveloop(timeout:integer);
var
  selectresult   : integer;
  fds            : fdset;
  endtime        : ttimeval;
  currenttimeout : ttimeval;
  selecttimeout	 : ttimeval;
  currenttime    : ttimeval;
  src            : tinetsockaddr;
  srclen         : integer;
begin
  gettimeofday(endtime);
  endtime.tv_sec := endtime.tv_sec + timeout;

  setupsocket;
  repeat
    state_process(state);
    case state.resultaction of
      action_ignore: begin
{        writeln('ignore');}
        {do nothing}
      end;
      action_done: begin
{        writeln('done');}
        exit;
        //onrequestdone(self,0);
      end;
      action_sendquery:begin
{        writeln('send query');}
        sendquery(state.packet,state.packetlen);
      end;
    end;
    gettimeofday(currenttime);
    selecttimeout := endtime;
    tv_substract(selecttimeout,currenttime);
    fd_zero(fds);
    fd_set(fd,fds);
    selectresult := select(fd+1,@fds,nil,nil,@selecttimeout);
    if selectresult > 0 then begin
{      writeln('selectresult>0');}
      fillchar(state.packet,sizeof(state.packet),0);
      state.packetlen := fdread(fd,state.packet, SizeOf(state.packet));
      //srclen := sizeof(src);
      //state.packetlen := sockets.recvfrom(fd,state.packet, SizeOf(state.packet),0,src,srclen);
      //writeln('packet received from: '+inttohex(htonl(src.addr),8)+':',htons(src.port));
      state.parsepacket := true;
    end;
  until selectresult <= 0;
end;

function forwardlookup(name:string;timeout:integer):tbinip;
begin
  ipstrtobin(name,result);
  if result.family <> 0 then exit;

  setstate_forward(name,state,0);
  resolveloop(timeout);
  result := state.resultbin;
end;

end.


